/**
   qdice - Copyright (C) 2006 - 2012 Simeon Simeonov

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

#include <QtGui>
#include <QtOpenGL>

#include "glwidget.hh"

/* coordinates */

#define X 0
#define Y 1
#define Z 2

#define X_LOWER -1
#define X_UPPER 1

#define Y_LOWER -1
#define Y_UPPER 1

#define Z_LOWER -1
#define Z_UPPER 1

#define OSD_TIMEOUT 3 /* osd timeout in seconds */


GLWidget::GLWidget(QWidget *parent) : QGLWidget(parent)
{
  
  d_object = 0;

  srand(time(0));

  turns = 0;
  sign = rangle = xr = yr = zr = 0;

  animation = false;
  display_help = false;

  interval = 80;

  osd_counter = OSD_TIMEOUT * 1000;

  /* the timer will call executeTicks() each 80 milliseconds */
  timer = new QTimer(this);
  connect(timer, SIGNAL(timeout()), this, SLOT(executeTicks()));
  timer->start(interval); 
    
}


GLWidget::~GLWidget()
{

  makeCurrent();
  glDeleteLists(d_object, 1);
    
}


QSize 
GLWidget::minimumSizeHint() const
{
  return QSize(80, 60);
}


QSize 
GLWidget::sizeHint() const
{
  return QSize(600, 600);
}


void GLWidget::initializeGL()
{

  /* set the clear-values for the color and the depth buffers */
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClearDepth(1.0);

  glShadeModel(GL_SMOOTH);

  glEnable(GL_DEPTH_TEST);


  /* no need to consider back-facing polygons */
  glEnable(GL_CULL_FACE);

  /* enable texturing */
  glEnable(GL_TEXTURE_2D);

  makeCurrent();

  d_object = makeDObject(); /* initialize the display list*/

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  /* does the same job as gluLookAt(0.0, 0.0, 10.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0); */
  glTranslated(0.0, 0.0, -10.0); 
  glGetDoublev(GL_MODELVIEW_MATRIX, tmp_matrix); /* load the current modelview matrix into tmp_matrix */

}


void GLWidget::paintGL()
{

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  glPushMatrix();

  if (display_help) /* whether to display the help message */
    displayHelp();

  if (osd_counter < OSD_TIMEOUT * 1000) {
    QFont serifFont("Times", 12);
    renderText (4.0, -3.0, -10.0, QString("Speed: %1").arg( (80 - interval) / 20 ), serifFont);
  }

  glLoadMatrixd(tmp_matrix); /* the modelview matrix is set to tmp_matrix */
  if (animation)
    glRotated( (GLdouble) sign * (GLdouble) rangle, 
			   (GLdouble) xr, 
			   (GLdouble) yr, 
			   (GLdouble) zr);

  if (animation && rangle >= 90) /* after each animation we update the tmp_matrix */
    glGetDoublev(GL_MODELVIEW_MATRIX, tmp_matrix);

  glCallList(d_object); /* draw the dice */


  glPopMatrix();

  swapBuffers();
  
}


void GLWidget::resizeGL(int width, int height)
{

  GLdouble aspect;

  glViewport(0, 0, width, height);
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  aspect = (GLdouble) width / (GLdouble) height;

  /* gluPerspective(60.0, (GLdouble) width / (GLdouble) height, 5.0, 60.0); */
  glFrustum(-2.0 * aspect, 2.0 * aspect, -2.0, 2.0, 5.0, 60.0); 

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
}


void
GLWidget::keyPressEvent (QKeyEvent * event)
{

  switch (event->key()) {

  case Qt::Key_Q:
    close();
    break;

  case Qt::Key_H:
    if (display_help)
      display_help = false;
    else
      display_help = true;
    break;

  case Qt::Key_Plus:
    if (interval > 20) {
      interval -= 20;
      timer->stop();
      timer->start(interval);
    }
    osd_counter = 0;
    break;

  case Qt::Key_Minus:
    if (interval < 160) {
      interval += 20;
      timer->stop();
      timer->start(interval);
    }
    osd_counter = 0;
    break;
    
  }

}


void 
GLWidget::mouseDoubleClickEvent(QMouseEvent *event)
{

  if (event->buttons() & Qt::LeftButton) {

    throwDice();

  }

}


void 
GLWidget::executeTicks()
{

  if (osd_counter < OSD_TIMEOUT * 1000)
    osd_counter += interval;

  if (animation) {

    if (turns <= 0) { /* animation completed */

      animation = false; 

    } else {

      if (rangle >= 90) { /* a whole step is completed */

        turns--;
        rangle = 0;
        makeRandomStep(); /* generate random data for a new step */

      } else {

        rangle += 18;

      }

    }
 
  }
    updateGL();

}


GLuint
GLWidget::makeDObject()
{

  int i, j;

  const int coords[6][4][3] = {
    { { X_UPPER, Y_LOWER, Z_LOWER }, { X_UPPER, Y_UPPER, Z_LOWER }, { X_UPPER, Y_UPPER, Z_UPPER }, { X_UPPER, Y_LOWER, Z_UPPER } }, /* right */
    { { X_LOWER, Y_LOWER, Z_UPPER }, { X_UPPER, Y_LOWER, Z_UPPER }, { X_UPPER, Y_UPPER, Z_UPPER }, { X_LOWER, Y_UPPER, Z_UPPER } }, /* front */
    { { X_LOWER, Y_UPPER, Z_LOWER }, { X_LOWER, Y_UPPER, Z_UPPER }, { X_UPPER, Y_UPPER, Z_UPPER }, { X_UPPER, Y_UPPER, Z_LOWER } }, /* top */
    { { X_LOWER, Y_LOWER, Z_LOWER }, { X_UPPER, Y_LOWER, Z_LOWER }, { X_UPPER, Y_LOWER, Z_UPPER }, { X_LOWER, Y_LOWER, Z_UPPER } }, /* bottom */
    { { X_LOWER, Y_LOWER, Z_LOWER }, { X_LOWER, Y_UPPER, Z_LOWER }, { X_UPPER, Y_UPPER, Z_LOWER }, { X_UPPER, Y_LOWER, Z_LOWER } }, /* back */
    { { X_LOWER, Y_LOWER, Z_LOWER }, { X_LOWER, Y_LOWER, Z_UPPER }, { X_LOWER, Y_UPPER, Z_UPPER }, { X_LOWER, Y_UPPER, Z_LOWER } } /* left */
  };


  GLuint textures[6];

  for (i = 0; i < 6; i++)
    textures[i] = bindTexture(QPixmap(QString("images/dice%1.png").arg(i + 1)), GL_TEXTURE_2D);

  GLuint list = glGenLists(1);
  glNewList(list, GL_COMPILE);

  glColor4f(1.0, 1.0, 1.0, 1.0);

  for (i = 0; i < 6; i++) {
    glBindTexture(GL_TEXTURE_2D, textures[i]);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

    glBegin(GL_QUADS);

    for (j = 0; j < 4; j++) {

      glTexCoord2d(j == 0 || j == 3, j == 0 || j == 1);
      glVertex3i(coords[i][j][0], coords[i][j][1], coords[i][j][2]);

    }
    glEnd();
  }

  glEndList();
  return list;

}


void 
GLWidget::throwDice()
{

  makeRandomStep();
  turns = rand() % 10 + 7;
  animation = true;

}


void 
GLWidget::makeRandomStep()
{

  int rnd_dir, rnd_sign;
  xr = yr = zr = 0;

  rnd_dir = rand() % 5; /* 2/5 chance for x-rotation, 2/5 for y-rotation, 1/5 for z-rotation */
  rnd_sign = rand() % 3; /* 2/3 chance to get a negative (CW) rotation */

  if (rnd_sign)
    sign = -1;
  else 
    sign = 1;

  if (rnd_dir == 0)
    zr = 1;
  else if ((rnd_dir == 1) || (rnd_dir == 2))
    xr = 1;
  else if ((rnd_dir == 3) || (rnd_dir == 4))
    yr = 1;

}


void
GLWidget::displayHelp()
{

  QFont serifFont("Times", 11);
  renderText (-5.5, 4.0, -10.0, "qdice-1.1", serifFont);
  renderText (-5.5, 3.0, -10.0, "Keyboard:", serifFont);
  renderText (-5.5, 2.5, -10.0, "+ - speed up the animation", serifFont);
  renderText (-5.5, 2.0, -10.0, "- - slow down the animation", serifFont);
  renderText (-5.5, 1.5, -10.0, "h - display this message (default - off)", serifFont);
  renderText (-5.5, 1.0, -10.0, "q - quit", serifFont);
  renderText (-5.5, -2.0, -10.0, "Mouse: ", serifFont);
  renderText (-5.5, -2.5, -10.0, "double click (left button) - throw the dice", serifFont);
  
}
